use anyhow::{bail, Context, Result};
use image::io::Reader;
use serde::Deserialize;
use std::{collections::HashMap, path::PathBuf};

use crate::{handle::TextureHandle, Content};

pub(crate) mod syntax {
	use serde::Deserialize;
	use std::path::PathBuf;

	use super::RepeatMode;

	// Raw serde syntax structs.
	// These are never seen by code outside this crate.

	#[derive(Debug, Deserialize)]
	#[serde(untagged)]
	pub enum Texture {
		Static(StaticTexture),
		Frames(FramesTexture),
	}

	#[derive(Debug, Deserialize)]
	pub struct StaticTexture {
		pub file: PathBuf,
	}

	#[derive(Debug, Deserialize)]
	pub struct FramesTexture {
		pub frames: Vec<PathBuf>,
		pub duration: f32,
		pub repeat: RepeatMode,
	}
}

/// How to replay a texture's animation
#[derive(Debug, Deserialize, Clone, Copy)]
pub enum RepeatMode {
	/// Play this animation once, and stop at the last frame
	#[serde(rename = "once")]
	Once,

	/// After the first frame, jump to the last frame
	#[serde(rename = "repeat")]
	Repeat,
}

impl RepeatMode {
	/// Represent this repeatmode as an integer
	/// Used to pass this enum into shaders
	pub fn as_int(&self) -> u32 {
		match self {
			Self::Once => 0,
			Self::Repeat => 1,
		}
	}
}

/// Represents a texture that may be used in the game.
#[derive(Debug, Clone)]
pub struct Texture {
	/// The name of this texture
	pub name: String,

	/// The handle for this texture
	pub handle: TextureHandle,

	/// The frames of this texture
	/// (static textures have one frame)
	pub frames: Vec<PathBuf>,

	/// The speed of this texture's animation
	/// (static textures have zero fps)
	pub fps: f32,

	/// How to replay this texture's animation
	pub repeat: RepeatMode,
}

impl crate::Build for Texture {
	type InputSyntax = HashMap<String, syntax::Texture>;

	fn build(texture: Self::InputSyntax, ct: &mut Content) -> Result<()> {
		for (texture_name, t) in texture {
			match t {
				syntax::Texture::Static(t) => {
					let file = ct.texture_root.join(t.file);
					let reader = Reader::open(&file).with_context(|| {
						format!(
							"Failed to read texture `{}` from file `{}`",
							texture_name,
							file.display()
						)
					})?;
					let dim = reader.into_dimensions().with_context(|| {
						format!(
							"Failed to get dimensions of texture `{}` from file `{}`",
							texture_name,
							file.display()
						)
					})?;

					let h = TextureHandle {
						index: ct.textures.len(),
						aspect: dim.0 as f32 / dim.1 as f32,
					};

					if texture_name == ct.starfield_texture_name {
						if ct.starfield_handle.is_none() {
							ct.starfield_handle = Some(h)
						} else {
							// This can't happen, since this is a hashmap.
							unreachable!("Found two starfield textures! Something is very wrong.")
						}
					}

					ct.texture_index.insert(texture_name.clone(), h);

					ct.textures.push(Self {
						name: texture_name,
						frames: vec![file],
						fps: 0.0,
						handle: h,
						repeat: RepeatMode::Once,
					});
				}
				syntax::Texture::Frames(t) => {
					let mut dim = None;
					for f in &t.frames {
						let file = ct.texture_root.join(f);
						let reader = Reader::open(&file).with_context(|| {
							format!(
								"Failed to read texture `{}` from file `{}`",
								texture_name,
								file.display()
							)
						})?;
						let d = reader.into_dimensions().with_context(|| {
							format!(
								"Failed to get dimensions of texture `{}` from file `{}`",
								texture_name,
								file.display()
							)
						})?;
						match dim {
							None => dim = Some(d),
							Some(e) => {
								if d != e {
									bail!(
										"Failed to load frames of texture `{}`. Frames have different sizes `{}`",
										texture_name,
										file.display()
									)
								}
							}
						}
					}

					let h = TextureHandle {
						index: ct.textures.len(),
						aspect: dim.unwrap().0 as f32 / dim.unwrap().1 as f32,
					};

					if texture_name == ct.starfield_texture_name {
						unreachable!("Starfield texture may not be animated")
					}

					let fps = t.duration / t.frames.len() as f32;

					ct.texture_index.insert(texture_name.clone(), h);
					ct.textures.push(Self {
						name: texture_name,
						frames: t
							.frames
							.into_iter()
							.map(|f| ct.texture_root.join(f))
							.collect(),
						fps,
						handle: h,
						repeat: t.repeat,
					});
				}
			}
		}

		if ct.starfield_handle.is_none() {
			bail!(
				"Could not find a starfield texture (name: `{}`)",
				ct.starfield_texture_name
			)
		}

		return Ok(());
	}
}
